{'config-search-paths': ['{:user-config-path}/sparcur/config.yaml',],
 'auth-variables':
 {'export-path': {'default': '{:user-data-path}/sparcur/export',
                  'environment-variables': 'SPARCUR_EXPORT_PATH SPARC_EXPORTS EXPORT_PATH'},
  'cache-path': {'default': '{:user-cache-path}/sparcur',
                 'environment-variables': 'SPARCUR_CACHE_PATH CACHE_PATH'},
  'log-path': {'default': '{:user-log-path}/sparcur',
               'environment-variables': 'SPARCUR_LOG_PATH LOG_PATH'},
  'blackfynn-organization': {'environment-variables': 'BLACKFYNN_ORGANIZATION'},
  'hypothesis-api-key': {'environment-variables': 'HYP_API_KEY HYP_API_TOKEN'},
  'hypothesis-group': {'environment-variables': 'HYP_GROUP'},
  'hypothesis-user': {'environment-variables': 'HYP_USER'},
  'datasets-noexport': None,
  'datasets-sparse': None,
  'datasets-no': None,
  'sparse-limit': {'default': 10000,
                   'environment-variables': 'SPARCUR_SPARSE_LIMIT SPARSE_LIMIT'},}}
